package dev.fyloz.colorrecipesexplorer.logic

import com.nhaarman.mockitokotlin2.*
import dev.fyloz.colorrecipesexplorer.model.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.TestInstance
import org.junit.jupiter.api.assertThrows
import kotlin.test.assertEquals
import kotlin.test.assertTrue

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
class InventoryLogicTest {
    private val materialLogic: MaterialLogic = mock()
    private val mixLogic: MixLogic = mock()
    private val logic = spy(DefaultInventoryLogic(materialLogic, mixLogic))

    @AfterEach
    fun afterEach() {
        reset(materialLogic, logic)
    }

    // add()

    @Test
    fun `add(materialQuantities) calls add() for each MaterialQuantityDto`() {
        val materialQuantities = listOf(
          materialQuantityDto(materialId = 1, quantity = 1234f),
          materialQuantityDto(materialId = 2, quantity = 2345f),
          materialQuantityDto(materialId = 3, quantity = 3456f),
          materialQuantityDto(materialId = 4, quantity = 4567f)
        )
        val storedQuantity = 2000f

        doAnswer { storedQuantity + (it.arguments[0] as MaterialQuantityDto).quantity }.whenever(logic)
          .add(any<MaterialQuantityDto>())

        val found = logic.add(materialQuantities)

        materialQuantities.forEach {
            verify(logic).add(it)
            assertTrue { found.any { updated -> updated.material == it.material && updated.quantity == storedQuantity + it.quantity } }
        }
    }

    @Test
    fun `add(materialQuantity) updates material's quantity`() {
        withGivenQuantities(0f, 1000f) {
            val updatedQuantity = it + this.quantity
            whenever(materialLogic.updateQuantity(any(), eq(this.quantity))).doReturn(updatedQuantity)

            val found = logic.add(this)

            verify(materialLogic).updateQuantity(
              argThat { this.id == this@withGivenQuantities.material },
              eq(this.quantity)
            )
            assertEquals(updatedQuantity, found)
        }
    }

    // deductMix()

    @Test
    fun `deductMix() calls deduct() with a collection of MaterialQuantityDto with adjusted quantities`() {
        val material = material(id = 0L, materialType = materialType(usePercentages = false))
        val materialPercents = material(id = 1L, materialType = materialType(usePercentages = true))
        val mixRatio = mixRatio(ratio = 1.5f)
        val mix = mix(
          id = mixRatio.id,
          mixMaterials = mutableSetOf(
            mixMaterial(id = 0L, material = material, quantity = 1000f, position = 0),
            mixMaterial(id = 1L, material = materialPercents, quantity = 50f, position = 1)
          )
        )
        val expectedQuantities = mapOf(
          0L to 1500f,
          1L to 750f
        )

        whenever(mixLogic.getById(mix.id!!)).doReturn(mix)
        doAnswer {
            (it.arguments[0] as Collection<MaterialQuantityDto>).map { materialQuantity ->
                materialQuantityDto(materialId = materialQuantity.material, quantity = 0f)
            }
        }.whenever(logic).deduct(any<Collection<MaterialQuantityDto>>())

        val found = logic.deductMix(mixRatio)

        verify(logic).deduct(argThat<Collection<MaterialQuantityDto>> {
            this.all { it.quantity == expectedQuantities[it.material] }
        })

        assertEquals(expectedQuantities.size, found.size)
    }

    // deduct()

    @Test
    fun `deduct(materialQuantities) calls deduct() for each MaterialQuantityDto`() {
        val materialQuantities = listOf(
          materialQuantityDto(materialId = 1, quantity = 1234f),
          materialQuantityDto(materialId = 2, quantity = 2345f),
          materialQuantityDto(materialId = 3, quantity = 3456f),
          materialQuantityDto(materialId = 4, quantity = 4567f)
        )
        val storedQuantity = 5000f

        doAnswer { storedQuantity - (it.arguments[0] as MaterialQuantityDto).quantity }.whenever(logic)
          .deduct(any<MaterialQuantityDto>())

        val found = logic.deduct(materialQuantities)

        materialQuantities.forEach {
            verify(logic).deduct(it)
            assertTrue { found.any { updated -> updated.material == it.material && updated.quantity == storedQuantity - it.quantity } }
        }
    }

    @Test
    fun `deduct(materialQuantities) throws MultiplesNotEnoughInventoryException when there is not enough inventory of a given material`() {
        val materialQuantities = listOf(
          materialQuantityDto(materialId = 1, quantity = 1234f),
          materialQuantityDto(materialId = 2, quantity = 2345f),
          materialQuantityDto(materialId = 3, quantity = 3456f),
          materialQuantityDto(materialId = 4, quantity = 4567f)
        )
        val inventoryQuantity = 3000f

        materialQuantities.forEach {
            withGivenQuantities(inventoryQuantity, it)
        }

        assertThrows<MultiplesNotEnoughInventoryException> { logic.deduct(materialQuantities) }
    }

    @Test
    fun `deduct(materialQuantity) updates material's quantity`() {
        withGivenQuantities(5000f, 1000f) {
            val updatedQuantity = it - this.quantity
            whenever(materialLogic.updateQuantity(any(), eq(-this.quantity))).doReturn(updatedQuantity)

            val found = logic.deduct(this)

            verify(materialLogic).updateQuantity(
              argThat { this.id == this@withGivenQuantities.material },
              eq(-this.quantity)
            )
            assertEquals(updatedQuantity, found)
        }
    }

    @Test
    fun `deduct(materialQuantity) throws NotEnoughInventoryException when there is not enough inventory of the given material`() {
        withGivenQuantities(0f, 1000f) {
            assertThrows<NotEnoughInventoryException> { logic.deduct(this) }
        }
    }

    private fun withGivenQuantities(
      stored: Float,
      quantity: Float,
      materialId: Long = 0L,
      test: MaterialQuantityDto.(Float) -> Unit = {}
    ) {
        val materialQuantity = materialQuantityDto(materialId = materialId, quantity = quantity)

        withGivenQuantities(stored, materialQuantity, test)
    }

    private fun withGivenQuantities(
      stored: Float,
      materialQuantity: MaterialQuantityDto,
      test: MaterialQuantityDto.(Float) -> Unit = {}
    ) {
        val material = material(id = materialQuantity.material, inventoryQuantity = stored)

        whenever(materialLogic.getById(material.id!!)).doReturn(material)

        materialQuantity.test(stored)
    }
}
