package dev.fyloz.colorrecipesexplorer.logic

import com.nhaarman.mockitokotlin2.*
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.logic.files.WriteableFileLogic
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.repository.MaterialRepository
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.TestInstance
import org.junit.jupiter.api.assertThrows
import org.springframework.mock.web.MockMultipartFile
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertTrue

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
class MaterialLogicTest :
    AbstractExternalNamedModelServiceTest<Material, MaterialSaveDto, MaterialUpdateDto, MaterialLogic, MaterialRepository>() {
    override val repository: MaterialRepository = mock()
    private val recipeService: RecipeLogic = mock()
    private val mixService: MixLogic = mock()
    private val materialTypeService: MaterialTypeLogic = mock()
    private val fileService: WriteableFileLogic = mock()
    override val logic: MaterialLogic =
        spy(DefaultMaterialLogic(repository, recipeService, mixService, materialTypeService, fileService, mock()))

    override val entity: Material = material(id = 0L, name = "material")
    private val entityOutput = materialOutputDto(entity)
    override val anotherEntity: Material = material(id = 1L, name = "another material")
    override val entityWithEntityName: Material = material(id = 2L, name = "material")
    override val entitySaveDto: MaterialSaveDto = spy(materialSaveDto())
    override val entityUpdateDto: MaterialUpdateDto = spy(materialUpdateDto(id = 0L))

    private val materialType = materialType()

    @AfterEach
    override fun afterEach() {
        reset(recipeService, mixService, materialTypeService, fileService)
        super.afterEach()
    }

    // existsByMaterialType

    @Test
    fun `existsByMaterialType() returns true when a material with the given material type exists in the repository`() {
        whenever(repository.existsByMaterialType(materialType)).doReturn(true)

        val found = logic.existsByMaterialType(materialType)

        assertTrue(found)
    }

    @Test
    fun `existsByMaterialType() returns false when no material with the given material type exists in the repository`() {
        whenever(repository.existsByMaterialType(materialType)).doReturn(false)

        val found = logic.existsByMaterialType(materialType)

        assertFalse(found)
    }

    // hasSimdut()

    @Test
    fun `hasSimdut() returns false when simdutService_exists() returns false`() {
        whenever(fileService.exists(any())).doReturn(false)
        doReturn(entity).whenever(logic).getById(entity.id!!)

        val found = logic.hasSimdut(entity)

        assertFalse(found)
    }

    @Test
    fun `hasSimdut() returns true when simdutService_exists() returns true`() {
        whenever(fileService.exists(any())).doReturn(true)
        doReturn(entity).whenever(logic).getById(entity.id!!)

        val found = logic.hasSimdut(entity)

        assertTrue(found)
    }

    // getAllNotMixType()

    @Test
    fun `getAllNotMixType() returns a list containing every material that are not a mix type`() {
        val mixTypeMaterial = material(id = 1L, name = "mix type material", isMixType = true)
        val mixTypeMaterialOutput = materialOutputDto(mixTypeMaterial)
        val materialList = listOf(entity, mixTypeMaterial)

        doReturn(materialList).whenever(logic).getAll()

        val found = logic.getAllNotMixType()

        assertTrue(found.contains(entityOutput))
        assertFalse(found.contains(mixTypeMaterialOutput))
    }

    // save()

    @Test
    fun `save() throws AlreadyExistsException when a material with the given name exists in the repository`() {
        doReturn(true).whenever(logic).existsByName(entity.name)

        assertThrows<AlreadyExistsException> { logic.save(entity) }
            .assertErrorCode("name")
    }

    @Test
    override fun `save(dto) calls and returns save() with the created entity`() {
        withBaseSaveDtoTest(entity, entitySaveDto, logic, { any() })
    }

    @Test
    fun `save(dto) calls simdutService_write() with the saved entity`() {
        val mockMultipartFile = spy(MockMultipartFile("simdut", byteArrayOf()))
        val materialSaveDto = spy(materialSaveDto(simdutFile = mockMultipartFile))

        doReturn(false).whenever(mockMultipartFile).isEmpty
        doReturn(entity).whenever(logic).save(any<Material>())

        logic.save(materialSaveDto)

        verify(fileService).write(mockMultipartFile, entity.simdutFilePath, false)
    }

    // update()

    @Test
    fun `update() throws AlreadyExistsException when another material with the updated name exists in the repository`() {
        val material = material(id = 0L, name = "name")
        val anotherMaterial = material(id = 1L, name = "name")

        whenever(repository.findByName(material.name)).doReturn(anotherMaterial)
        doReturn(entity).whenever(logic).getById(material.id!!)

        assertThrows<AlreadyExistsException> { logic.update(material) }
            .assertErrorCode("name")
    }

    @Test
    override fun `update(dto) calls and returns update() with the created entity`() {
        val mockSimdutFile = MockMultipartFile("simdut", byteArrayOf(1, 2, 3, 4, 5))
        val materialUpdateDto = spy(materialUpdateDto(id = 0L, simdutFile = mockSimdutFile))

        doReturn(entity).whenever(logic).getById(any())
        doReturn(entity).whenever(logic).update(any<Material>())
        doReturn(entity).whenever(materialUpdateDto).toEntity()

        logic.update(materialUpdateDto)

        verify(fileService).write(mockSimdutFile, entity.simdutFilePath, true)
    }

    // updateQuantity()

    @Test
    fun `updateQuantity() updates the quantity of the the given material in the repository`() {
        val material = material(id = 0L, inventoryQuantity = 4321f)
        val quantity = 1234f
        val totalQuantity = material.inventoryQuantity + quantity

        val found = logic.updateQuantity(material, quantity)

        verify(repository).updateInventoryQuantityById(material.id!!, totalQuantity)
        assertEquals(totalQuantity, found)
    }

    // getAllForMixCreation()

    @Test
    fun `getAllForMixCreation() returns all normal materials and all mix type materials for the given recipe`() {
        val normalMaterial = material(id = 0L, isMixType = false)
        val mixTypeMaterial = material(id = 1L, isMixType = true)
        val anotherMixTypeMaterial = material(id = 2L, isMixType = true)
        val materials = listOf(normalMaterial, mixTypeMaterial, anotherMixTypeMaterial)
        val recipe =
            recipe(id = 0L, mixes = mutableListOf(mix(mixType = mixType(id = 0L, material = mixTypeMaterial))))

        whenever(recipeService.getById(recipe.id!!)).doReturn(recipe)
        doReturn(materials).whenever(logic).getAll()

        val found = logic.getAllForMixCreation(recipe.id!!)

        assertTrue(materialOutputDto(normalMaterial) in found)
        assertTrue(materialOutputDto(mixTypeMaterial) in found)
        assertFalse(materialOutputDto(anotherMixTypeMaterial) in found)
    }

    // getAllForMixUpdate()

    @Test
    fun `getAllForMixUpdate() returns all normal materials and all mix type materials for the recipe of the given mix without the mix type of the said mix`() {
        val normalMaterial = material(id = 0L, isMixType = false)
        val mixTypeMaterial = material(id = 1L, isMixType = true)
        val anotherMixTypeMaterial = material(id = 2L, isMixType = true)
        val materials = listOf(normalMaterial, mixTypeMaterial, anotherMixTypeMaterial)
        val recipe = recipe(id = 0L, mixes = mutableListOf(mix(mixType = mixType(material = mixTypeMaterial))))
        val mix = mix(id = 1L, recipe = recipe, mixType = mixType(material = anotherMixTypeMaterial))
        recipe.mixes.add(mix)

        whenever(mixService.getById(mix.id!!)).doReturn(mix)
        doReturn(materials).whenever(logic).getAll()

        val found = logic.getAllForMixUpdate(mix.id!!)

        assertTrue(materialOutputDto(normalMaterial) in found)
        assertTrue(materialOutputDto(mixTypeMaterial) in found)
        assertFalse(materialOutputDto(anotherMixTypeMaterial) in found)
    }


    // delete()

    override fun `delete() deletes in the repository`() {
        whenCanBeDeleted {
            super.`delete() deletes in the repository`()
        }
    }

    // deleteById()

    override fun `deleteById() deletes the entity with the given id in the repository`() {
        whenCanBeDeleted {
            super.`deleteById() deletes the entity with the given id in the repository`()
        }
    }

    /** Utility property to check if the identifier of the given [Material] is even. */
    private val Material.evenId: Boolean
        get() = this.id!! % 2 == 0L

    private fun whenCanBeDeleted(id: Long = any(), test: () -> Unit) {
        whenever(repository.canBeDeleted(id)).doReturn(true)

        test()
    }

    private fun materialOutputDto(material: Material) = MaterialOutputDto(
        id = material.id!!,
        name = material.name,
        inventoryQuantity = material.inventoryQuantity,
        isMixType = material.isMixType,
        materialType = material.materialType!!,
        simdutUrl = null
    )
}
